/* Copyright (c) 2006 Dirk Jagdmann <doj@cubic.org>

This software is provided 'as-is', without any express or implied
warranty. In no event will the authors be held liable for any damages
arising from the use of this software.

Permission is granted to anyone to use this software for any purpose,
including commercial applications, and to alter it and redistribute it
freely, subject to the following restrictions:

    1. The origin of this software must not be misrepresented; you
       must not claim that you wrote the original software. If you use
       this software in a product, an acknowledgment in the product
       documentation would be appreciated but is not required.

    2. Altered source versions must be plainly marked as such, and
       must not be misrepresented as being the original software.

    3. This notice may not be removed or altered from any source
       distribution. */

/* $Header: /code/qtpod/alsamidi.cpp,v 1.7 2006/07/31 10:34:56 doj Exp $ */

#include <iostream>
#include <alsa/asoundlib.h>
#include "midi.hpp"
#include "qtpod.hpp"
#include <qbuttongroup.h>
#include <vector>
using namespace std;

static snd_seq_t *seq=0;
static int /*client,*/ clientId, port, ch;
static bool midiblock=false;

static void midi_close()
{
  if(seq)
    snd_seq_close(seq), seq=0;
}

void do_controller(int c, int v)
{
#ifdef DOJDEBUG
  if(c==30 || c==62) // to debug delay
  fprintf(stderr, "do_controller(%i,%i) %i\n",c,v, midiblock);
#endif
  if(midiblock) return;
  snd_seq_event_t ev;
  snd_seq_ev_clear(&ev);
  snd_seq_ev_set_source(&ev, port);
  snd_seq_ev_set_subs(&ev);
  snd_seq_ev_set_direct(&ev);
  snd_seq_ev_set_controller(&ev,ch,c,v);
  snd_seq_event_output(seq, &ev);
  snd_seq_drain_output(seq);
}

void do_programchange(int v)
{
#ifdef DOJDEBUG
  fprintf(stderr, "do_programchange(%i) %i\n",v, midiblock);
#endif
  if(midiblock) return;
  snd_seq_event_t ev;
  snd_seq_ev_clear(&ev);
  snd_seq_ev_set_source(&ev, port);
  snd_seq_ev_set_subs(&ev);
  snd_seq_ev_set_direct(&ev);
  snd_seq_ev_set_pgmchange(&ev,ch,v);
  snd_seq_event_output(seq, &ev);
  snd_seq_drain_output(seq);
}

void do_sysex(uint8_t *buf, unsigned len)
{
#ifdef DOJDEBUG
  fprintf(stderr, "do_sysex(%p,%i) %i\n", buf,len,midiblock);
#endif
  if(midiblock) return;
  snd_seq_event_t ev;
  snd_seq_ev_clear(&ev);
  snd_seq_ev_set_source(&ev, port);
  snd_seq_ev_set_subs(&ev);
  snd_seq_ev_set_direct(&ev);
  snd_seq_ev_set_sysex(&ev, len, buf);
  snd_seq_event_output(seq, &ev);
  snd_seq_drain_output(seq);
}

int midi_init(char *appname, int channel, char *midiclient)
{
  if(seq)
    return 0;

  atexit(midi_close);

  ch=channel;

  if(snd_seq_open(&seq, "default", SND_SEQ_OPEN_DUPLEX, SND_SEQ_NONBLOCK) < 0)
    {
      cerr << "could not init ALSA sequencer\n";
      seq=0;
      return -1;
    }

  snd_seq_set_client_name(seq, appname);
  snd_seq_client_info_t *clientinfo;
  snd_seq_client_info_alloca(&clientinfo);
  snd_seq_get_client_info (seq, clientinfo);
  //client=snd_seq_client_info_get_client(clientinfo);
  clientId = snd_seq_client_id(seq);

  port=snd_seq_create_simple_port(seq, "control port",
				  SND_SEQ_PORT_CAP_WRITE|SND_SEQ_PORT_CAP_SUBS_WRITE|SND_SEQ_PORT_CAP_READ|SND_SEQ_PORT_CAP_SUBS_READ,
				  SND_SEQ_PORT_TYPE_APPLICATION);
  if(port < 0)
    {
      cerr << "could not create ALSA sequencer port\n";
      return -1;
    }

  const int npfd=snd_seq_poll_descriptors_count(seq, POLLIN);
  if(npfd<=0)
    {
      cerr << "could not get number of ALSA sequencer poll descriptors\n";
      return -1;
    }

  struct pollfd *pfd=(struct pollfd*)alloca(npfd * sizeof(struct pollfd));
  if(pfd==0)
    {
      cerr << "could not alloc memory for ALSA sequencer poll descriptors\n";
      return -1;
    }
  if(snd_seq_poll_descriptors(seq, pfd, npfd, POLLIN) != npfd)
    {
      cerr << "number of returned poll desc is not equal of request poll desc\n";
      return -1;
    }

  if(midiclient)
    {
      char *p_=strchr(midiclient, ':');
      if(p_)
	{
	  *p_++=0;
	  const int p=strtoul(p_,0,0);
	  const int c=strtoul(midiclient,0,0);

	  snd_seq_addr_t sender, dest;
	  sender.client = clientId;
	  sender.port = port;
	  dest.client = c;
	  dest.port = p;
	  snd_seq_port_subscribe_t *subs;
	  snd_seq_port_subscribe_alloca(&subs);
	  snd_seq_port_subscribe_set_sender(subs, &sender);
	  snd_seq_port_subscribe_set_dest(subs, &dest);
	  snd_seq_subscribe_port(seq, subs);

	  sender.client = c;
	  sender.port = p;
	  dest.client = clientId;
	  dest.port = port;
	  snd_seq_port_subscribe_set_sender(subs, &sender);
	  snd_seq_port_subscribe_set_dest(subs, &dest);
	  snd_seq_subscribe_port(seq, subs);
	}
    }

  return pfd[0].fd;
}

void QTPOD::midievent(int)
{
  bool sendpq=false;
  midiblock=true;
  snd_seq_event_t *ev;
  do
    {
      snd_seq_event_input(seq, &ev);
      if(!ev) continue;
      switch(ev->type)
	{
	case SND_SEQ_EVENT_CONTROLLER:
#if 0
	  fprintf(stderr, "Channel %02d: Controller: %03d Value:%d\n",
		  ev->data.control.channel, ev->data.control.param, ev->data.control.value);
#endif
	  if(ev->data.control.channel == ch)
	    parsecontroller(ev->data.control.param, ev->data.control.value);
	  break;

	case SND_SEQ_EVENT_PGMCHANGE:
#if 0
	  fprintf(stderr, "Channel %02d: Program: %03d\n",
		  ev->data.control.channel, ev->data.control.value);
#endif
	  if(ev->data.control.channel == ch)
	    {
	      bg->setButton(ev->data.control.value);
	      sendpq=true;
	    }
	  break;

	case SND_SEQ_EVENT_PORT_SUBSCRIBED:
#if 0
	  fprintf(stderr, "event subscribed send.client:%i dest.client:%i clientId:%i\n",
		  (int)ev->data.connect.sender.client, (int)ev->data.connect.dest.client, clientId);
#endif
	  sendpq=true;
	  break;

	case SND_SEQ_EVENT_SYSEX:
	  if(ev->data.ext.len > 0 && ev->data.ext.ptr)
	    {
	      static vector<uint8_t> sysexbuf;
	      for(unsigned i=0; i<ev->data.ext.len; ++i)
		sysexbuf.push_back(reinterpret_cast<uint8_t*>(ev->data.ext.ptr)[i]);
	      if(sysexbuf[sysexbuf.size()-1] == 0xF7)
		{
		  parsesysex(&sysexbuf[0], sysexbuf.size());
		  sysexbuf.clear();
		}
	    }
	  break;
	}
      snd_seq_free_event(ev);
    }
  while (snd_seq_event_input_pending(seq, 0) > 0);
  midiblock=false;
  if(sendpq)
    sendpatchquery();
}
